# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from apitools.base.py import encoding
import mock
import os
import tempfile
import unittest2
import webtest
from expects import be_false, be_none, be_true, expect, equal, raise_error

from endpoints_management.auth import suppliers
from endpoints_management.auth import tokens
from endpoints_management.control import (client, report_request, service,
                                          sc_messages, sm_messages, wsgi)


def _dummy_start_response(status, response_headers, exc_info=None):
    pass


_DUMMY_RESPONSE = (b'This is the dummy app response.',)


class _DummyWsgiApp(object):
    # TODO: replace this with a response object from webob
    def __call__(self, environ, dummy_start_response):
        dummy_start_response("200 OK", [])
        return _DUMMY_RESPONSE


class TestEnvironmentMiddleware(unittest2.TestCase):

    def test_should_add_service_et_al_to_environment(self):
        cls = wsgi.EnvironmentMiddleware
        wrappee = _DummyWsgiApp()
        wanted_service = service.Loaders.SIMPLE.load()
        wrapped = cls(wrappee, wanted_service)

        given = {
            u'wsgi.url_scheme': u'http',
            u'HTTP_HOST': u'localhost',
            u'REQUEST_METHOD': u'GET'}
        wrapped(given, _dummy_start_response)
        expect(given.get(cls.SERVICE)).to(equal(wanted_service))
        expect(given.get(cls.SERVICE_NAME)).to(equal(wanted_service.name))
        expect(given.get(cls.METHOD_REGISTRY)).not_to(be_none)
        expect(given.get(cls.REPORTING_RULES)).not_to(be_none)
        expect(given.get(cls.METHOD_INFO)).not_to(be_none)

    def test_should_handle_method_override(self):
        cls = wsgi.EnvironmentMiddleware
        wrappee = _DummyWsgiApp()
        wanted_service = service.Loaders.SIMPLE.load()
        wrapped = cls(wrappee, wanted_service)

        given = {
            u'wsgi.url_scheme': u'http',
            u'HTTP_HOST': u'localhost',
            u'REQUEST_METHOD': u'POST',
            u'HTTP_X_HTTP_METHOD_OVERRIDE': u'PATCH',
        }

        wrapped(given, _dummy_start_response)
        assert given[cls.METHOD_INFO].selector == 'allow-all.PATCH'


class TestMiddleware(unittest2.TestCase):
    PROJECT_ID = u'middleware'

    def test_should_not_send_requests_if_there_is_no_service(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)

        given = {
            u'wsgi.url_scheme': u'http',
            u'PATH_INFO': u'/any',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId=u'fake_operation_id')
        wrapped = wsgi.Middleware(wrappee, self.PROJECT_ID, control_client)
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_false)
        expect(control_client.report.called).to(be_false)
        expect(control_client.allocate_quota.called).to(be_false)

    def test_should_send_requests_using_the_client(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)

        given = {
            u'wsgi.url_scheme': u'http',
            u'PATH_INFO': u'/any',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId=u'fake_operation_id')
        with_control = wsgi.Middleware(wrappee, self.PROJECT_ID, control_client)
        wrapped = wsgi.EnvironmentMiddleware(with_control,
                                             service.Loaders.SIMPLE.load())
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        expect(control_client.report.called).to(be_true)
        # no quota definitions in this service config
        expect(control_client.allocate_quota.called).to(be_false)

    def test_should_send_report_request_if_check_fails(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            u'wsgi.url_scheme': u'http',
            u'PATH_INFO': u'/any',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId = u'fake_operation_id',
            checkErrors = [
                sc_messages.CheckError(
                    code=sc_messages.CheckError.CodeValueValuesEnum.PROJECT_DELETED)
            ]
        )
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.SIMPLE)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        expect(control_client.report.called).to(be_true)
        expect(control_client.allocate_quota.called).to(be_false)

    def test_load_service_failed_retrying(self):
        control_client = mock.MagicMock(spec=client.Client)

        given = {
            u'wsgi.url_scheme': u'http',
            u'PATH_INFO': u'/any',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId=u'fake_operation_id')
        control_client.check.return_value = dummy_response

        loader = mock.MagicMock()
        # fail to load twice, then load
        loader.load.side_effect = [None, None, service.Loaders.SIMPLE.load()]
        # will consume first value
        result = wsgi.ConfigFetchWrapper(
            _DummyWsgiApp(), self.PROJECT_ID, control_client,
            loader=loader, disable_threading=True)

        test_app = webtest.TestApp(result)
        # will retry, consuming second value
        resp = test_app.get('/any', expect_errors=True)
        assert resp.status_code == 503
        # will retry, consuming third value
        resp = test_app.get('/any')
        assert resp.status_code == 200



_SYSTEM_PARAMETER_CONFIG_TEST = b"""
{
    "name": "system-parameter-config",
    "systemParameters": {
       "rules": [{
         "selector": "Uvw.Method1",
         "parameters": [{
            "name": "name1",
            "httpHeader": "Header-Key1",
            "urlQueryParameter": "param_key1"
         }, {
            "name": "name2",
            "httpHeader": "Header-Key2",
            "urlQueryParameter": "param_key2"
         }, {
            "name": "api_key",
            "httpHeader": "ApiKeyHeader",
            "urlQueryParameter": "ApiKeyParam"
         }, {
            "httpHeader": "Ignored-NoName-Key3",
            "urlQueryParameter": "Ignored-NoName-key3"
         }]
       }, {
         "selector": "Uvw.Method2",
         "parameters": [{
            "name": "name1",
            "httpHeader": "Header-Key1",
            "urlQueryParameter": "param_key1"
         }, {
            "name": "name2",
            "httpHeader": "Header-Key2",
            "urlQueryParameter": "param_key2"
         }, {
            "name": "api_key",
            "httpHeader": "ApiKeyHeader",
            "urlQueryParameter": "ApiKeyParam"
         }, {
            "httpHeader": "Ignored-NoName-Key3",
            "urlQueryParameter": "Ignored-NoName-key3"
         }]
       }, {
         "selector": "Bad.NotConfigured",
         "parameters": [{
            "name": "neverUsed",
            "httpHeader": "NeverUsed-Key1",
            "urlQueryParameter": "NeverUsed_key1"
         }]
       }]
    },
    "http": {
        "rules": [{
            "selector": "Uvw.Method1",
            "get": "/uvw/method1/{x}"
        }, {
            "selector": "Uvw.Method2",
            "get": "/uvw/method2/{x}"
        }, {
            "selector": "Uvw.MethodNeedsApiKey",
            "get": "/uvw/method_needs_api_key/{x}"
        }, {
            "selector": "Uvw.DefaultParameters",
            "get": "/uvw/default_parameters"
        }]
    },
    "metrics": [{
            "metricKind": "GAUGE",
            "name": "example-read-requests",
            "valueType": "INT64"
        }, {
            "metricKind": "GAUGE",
            "name": "example-list-requests",
            "valueType": "INT64"
        }
    ],
    "quota": {
        "limits": [{
            "displayName": "My Read API Requests per Minute",
            "metric": "example-read-requests",
            "name": "example-read-requests",
            "unit": "1/min/{project}",
            "values": {"STANDARD": "1000"}
        }, {
            "displayName": "My List API Requests per Minute",
            "metric": "example-list-requests",
            "name": "example-list-requests",
            "unit": "1/min/{project}",
            "values": {"STANDARD": "100"}
        }],
        "metricRules": [{
            "metricCosts": {
                 "example-list-requests": "1",
                 "example-read-requests": "5"
             },
            "selector": "Uvw.Method2"
        }]
    },
    "usage": {
        "rules": [{
            "selector" : "Uvw.Method1",
            "allowUnregisteredCalls" : true
        },  {
            "selector" : "Uvw.Method2",
            "allowUnregisteredCalls" : true
        },  {
            "selector": "Uvw.MethodNeedsApiKey",
            "allowUnregisteredCalls" : false
        }, {
            "selector" : "Uvw.DefaultParameters",
            "allowUnregisteredCalls" : true
        }]
    }
}
"""

class TestMiddlewareWithParams(unittest2.TestCase):
    PROJECT_ID = u'middleware-with-params'

    def setUp(self):
        _config_fd = tempfile.NamedTemporaryFile(delete=False)
        with _config_fd as f:
            f.write(_SYSTEM_PARAMETER_CONFIG_TEST)
        self._config_file = _config_fd.name
        os.environ[service.CONFIG_VAR] = self._config_file

    def tearDown(self):
        if os.path.exists(self._config_file):
            os.remove(self._config_file)

    def test_should_send_requests_with_no_param(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            u'wsgi.url_scheme': u'http',
            u'PATH_INFO': u'/uvw/method1/with_no_param',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId=u'fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        req = control_client.check.call_args[0][0]
        expect(req.checkRequest.operation.consumerId).to(
            equal(u'project:middleware-with-params'))
        expect(control_client.report.called).to(be_true)
        expect(control_client.allocate_quota.called).to(be_false)

    def test_should_send_quota_requests_with_no_param(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            u'wsgi.url_scheme': u'http',
            u'PATH_INFO': u'/uvw/method2/with_no_param',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId=u'fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        control_client.allocate_quota.side_effect = lambda req: sc_messages.AllocateQuotaResponse(
            operationId=req.allocateQuotaRequest.allocateOperation.operationId)
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        req = control_client.check.call_args[0][0]
        expect(req.checkRequest.operation.consumerId).to(
            equal(u'project:middleware-with-params'))
        expect(control_client.report.called).to(be_true)
        expect(control_client.allocate_quota.called).to(be_true)

    def test_should_send_requests_with_configured_query_param_api_key(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            u'wsgi.url_scheme': u'http',
            u'QUERY_STRING': u'ApiKeyParam=my-query-value',
            u'PATH_INFO': u'/uvw/method1/with_query_param',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId=u'fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        check_req = control_client.check.call_args[0][0]
        expect(check_req.checkRequest.operation.consumerId).to(
            equal(u'api_key:my-query-value'))
        expect(control_client.report.called).to(be_true)
        report_req = control_client.report.call_args[0][0]
        expect(report_req.reportRequest.operations[0].consumerId).to(
            equal(u'api_key:my-query-value'))
        expect(control_client.allocate_quota.called).to(be_false)

    def test_should_send_requests_with_configured_header_api_key(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            u'wsgi.url_scheme': u'http',
            u'PATH_INFO': u'/uvw/method1/with_query_param',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_APIKEYHEADER': u'my-header-value',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId=u'fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_true)
        check_request = control_client.check.call_args_list[0].checkRequest
        check_req = control_client.check.call_args[0][0]
        expect(check_req.checkRequest.operation.consumerId).to(
            equal(u'api_key:my-header-value'))
        expect(control_client.report.called).to(be_true)
        report_req = control_client.report.call_args[0][0]
        expect(report_req.reportRequest.operations[0].consumerId).to(
            equal(u'api_key:my-header-value'))
        expect(control_client.allocate_quota.called).to(be_false)

    def test_should_send_requests_with_default_query_param_api_key(self):
        for default_key in (u'key', u'api_key'):
            wrappee = _DummyWsgiApp()
            control_client = mock.MagicMock(spec=client.Client)
            given = {
                u'wsgi.url_scheme': u'http',
                u'QUERY_STRING': u'%s=my-default-api-key-value' % (default_key,),
                u'PATH_INFO': u'/uvw/method_needs_api_key/with_query_param',
                u'REMOTE_ADDR': u'192.168.0.3',
                u'HTTP_HOST': u'localhost',
                u'HTTP_REFERER': u'example.myreferer.com',
                u'REQUEST_METHOD': u'GET'}
            dummy_response = sc_messages.CheckResponse(
                operationId=u'fake_operation_id')
            wrapped = wsgi.add_all(wrappee,
                                   self.PROJECT_ID,
                                   control_client,
                                   loader=service.Loaders.ENVIRONMENT)
            control_client.check.return_value = dummy_response
            wrapped(given, _dummy_start_response)
            expect(control_client.check.called).to(be_true)
            check_request = control_client.check.call_args_list[0].checkRequest
            check_req = control_client.check.call_args[0][0]
            expect(check_req.checkRequest.operation.consumerId).to(
                equal(u'api_key:my-default-api-key-value'))
            expect(control_client.report.called).to(be_true)
            report_req = control_client.report.call_args[0][0]
            expect(report_req.reportRequest.operations[0].consumerId).to(
                equal(u'api_key:my-default-api-key-value'))
            expect(control_client.allocate_quota.called).to(be_false)

    def test_should_not_perform_check_if_needed_api_key_is_missing(self):
        wrappee = _DummyWsgiApp()
        control_client = mock.MagicMock(spec=client.Client)
        given = {
            u'wsgi.url_scheme': u'http',
            u'PATH_INFO': u'/uvw/method_needs_api_key/more_stuff',
            u'REMOTE_ADDR': u'192.168.0.3',
            u'HTTP_HOST': u'localhost',
            u'HTTP_REFERER': u'example.myreferer.com',
            u'REQUEST_METHOD': u'GET'}
        dummy_response = sc_messages.CheckResponse(
            operationId=u'fake_operation_id')
        wrapped = wsgi.add_all(wrappee,
                               self.PROJECT_ID,
                               control_client,
                               loader=service.Loaders.ENVIRONMENT)
        control_client.check.return_value = dummy_response
        wrapped(given, _dummy_start_response)
        expect(control_client.check.called).to(be_false)
        expect(control_client.report.called).to(be_true)
        report_req = control_client.report.call_args[0][0]
        expect(report_req.reportRequest.operations[0].consumerId).to(
            equal(u'project:middleware-with-params'))
        expect(control_client.allocate_quota.called).to(be_false)

AuthMiddleware = wsgi.AuthenticationMiddleware


class TestAuthenticationMiddleware(unittest2.TestCase):

    def setUp(self):
        self._mock_application = _DummyWsgiApp()
        self._mock_authenticator = mock.MagicMock(spec=tokens.Authenticator)
        self._middleware = AuthMiddleware(self._mock_application,
                                          self._mock_authenticator)

    def test_no_authentication(self):
        with self.assertRaisesRegex(ValueError, u"Invalid authenticator"):
            AuthMiddleware(self._mock_application, None)

    def test_no_method_info(self):
        environ = {}
        self.assertEqual(_DUMMY_RESPONSE,
                         self._middleware(environ, _dummy_start_response))

    def test_no_auth_token(self):
        auth_app = AuthMiddleware(self.UserInfoWsgiApp(), self._mock_authenticator)
        method_info = mock.MagicMock()
        method_info.auth_info = mock.MagicMock()
        environ = {
            wsgi.EnvironmentMiddleware.METHOD_INFO: method_info
        }
        self.assertIsNone(auth_app(environ, _dummy_start_response))

    def test_malformed_authorization_header(self):
        auth_app = AuthMiddleware(self.UserInfoWsgiApp(), self._mock_authenticator)
        environ = {
            u"HTTP_AUTHORIZATION": u"malformed-auth-token",
            wsgi.EnvironmentMiddleware.METHOD_INFO: mock.MagicMock(),
            wsgi.EnvironmentMiddleware.SERVICE_NAME: u"service-name"}
        self._mock_authenticator.authenticate.side_effect = suppliers.UnauthenticatedException()
        self.assertIsNone(auth_app(environ, _dummy_start_response))

    def test_successful_authentication(self):
        auth_token = u"Bearer test-bearer-token"
        auth_info = mock.MagicMock()
        service_name = u"test-service-name"
        method_info = mock.MagicMock()
        method_info.auth_info = auth_info
        environ = {
            u"HTTP_AUTHORIZATION": auth_token,
            wsgi.EnvironmentMiddleware.METHOD_INFO: method_info,
            wsgi.EnvironmentMiddleware.SERVICE_NAME: service_name
        }

        user_info = mock.MagicMock()
        self._mock_authenticator.authenticate.return_value = user_info
        self._middleware(environ, _dummy_start_response)
        self.assertEqual(user_info, environ.get(AuthMiddleware.USER_INFO))
        authenticate_mock = self._mock_authenticator.authenticate
        authenticate_mock.assert_called_once_with(u"test-bearer-token", auth_info,
                                                  service_name)

    def test_auth_token_in_query(self):
        auth_token = u"test-bearer-token"
        auth_info = mock.MagicMock()
        service_name = u"test-service-name"
        method_info = mock.MagicMock()
        method_info.auth_info = auth_info
        environ = {
            u"QUERY_STRING": u"access_token=" + auth_token,
            wsgi.EnvironmentMiddleware.METHOD_INFO: method_info,
            wsgi.EnvironmentMiddleware.SERVICE_NAME: service_name
        }

        user_info = mock.MagicMock()
        self._mock_authenticator.authenticate.return_value = user_info
        self._middleware(environ, _dummy_start_response)
        self.assertEqual(user_info, environ.get(AuthMiddleware.USER_INFO))
        authenticate_mock = self._mock_authenticator.authenticate
        authenticate_mock.assert_called_once_with(u"test-bearer-token", auth_info,
                                                  service_name)

    patched_environ = {}
    @mock.patch(u"os.environ", patched_environ)
    def test_set_user_info(self):
        environ = {
            u"QUERY_STRING": u"access_token=test-token",
            wsgi.EnvironmentMiddleware.METHOD_INFO: mock.MagicMock(),
            wsgi.EnvironmentMiddleware.SERVICE_NAME: u"test-service-name"}
        application = self.UserInfoWsgiApp()
        auth_middleware = AuthMiddleware(application, self._mock_authenticator)
        user_info = mock.MagicMock()
        self._mock_authenticator.authenticate.return_value = user_info
        self.assertEqual(user_info, auth_middleware(environ,
                                                    _dummy_start_response))
        self.assertFalse(self.patched_environ)

    class UserInfoWsgiApp(object):
        def __call__(self, environ, start_response):
            return os.environ.get(wsgi.AuthenticationMiddleware.USER_INFO)


class TestCreateAuthenticator(unittest2.TestCase):
    def test_create_without_service(self):
        with self.assertRaises(ValueError):
            wsgi._create_authenticator(None)

    def test_load_service_without_auth(self):
        service = _read_service_from_json(u"{}")
        self.assertIsNone(wsgi._create_authenticator(service))

    def test_load_service(self):
        json = """{
            "authentication": {
                "providers": [{
                    "issuer": "auth-issuer"
                }]
            }
        }"""
        service = _read_service_from_json(json)
        self.assertIsNotNone(wsgi._create_authenticator(service))


patched_platform_environ = {}
@mock.patch.dict(u'os.environ', patched_platform_environ, clear=True)
class TestPlatformDetection(unittest2.TestCase):

    def test_development(self):
        os.environ[u'SERVER_SOFTWARE'] = u'Development/2.0.0'
        self.assertEqual(report_request.ReportedPlatforms.DEVELOPMENT,
                         wsgi._get_platform())

    def test_gke(self):
        os.environ[u'KUBERNETES_SERVICE_HOST'] = u'hostname'
        self.assertEqual(report_request.ReportedPlatforms.GKE,
                         wsgi._get_platform())

    @mock.patch.object(wsgi, u'_running_on_gce', return_value=True)
    def test_gae_flex(self, _running_on_gce):
        os.environ[u'GAE_MODULE_NAME'] = u'gae_module'
        self.assertEqual(report_request.ReportedPlatforms.GAE_FLEX,
                         wsgi._get_platform())

    @mock.patch.object(wsgi, u'_running_on_gce', return_value=True)
    def test_gce(self, _running_on_gce):
        self.assertEqual(report_request.ReportedPlatforms.GCE,
                         wsgi._get_platform())

    @mock.patch.object(wsgi, u'_running_on_gce', return_value=False)
    def test_gae_standard(self, _running_on_gce):
        os.environ[u'SERVER_SOFTWARE'] = u'Google App Engine/1.2.3'
        self.assertEqual(report_request.ReportedPlatforms.GAE_STANDARD,
                         wsgi._get_platform())

    @mock.patch.object(wsgi, u'_running_on_gce', return_value=False)
    def test_unknown(self, _running_on_gce):
        self.assertEqual(report_request.ReportedPlatforms.UNKNOWN,
                         wsgi._get_platform())


def _read_service_from_json(json):
    return encoding.JsonToMessage(sm_messages.Service, json)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime
import httplib
import unittest2
import mock
from operator import attrgetter
from expects import be, be_none, be_true, be_false, equal, expect, raise_error

from apitools.base.py import encoding

from endpoints_management.control import caches, label_descriptor, timestamp
from endpoints_management.control import (quota_request, metric_value,
                                          sc_messages)


class TestSign(unittest2.TestCase):

    def setUp(self):
        op = sc_messages.QuotaOperation(
            consumerId=_TEST_CONSUMER_ID,
            methodName=_TEST_OP_NAME
        )
        self.test_quota_request = sc_messages.AllocateQuotaRequest(allocateOperation=op)
        self.test_op = op

    def test_should_fail_if_operation_is_not_set(self):
        testf = lambda: quota_request.sign(sc_messages.AllocateQuotaRequest())
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_on_invalid_input(self):
        testf = lambda: quota_request.sign(None)
        expect(testf).to(raise_error(ValueError))
        testf = lambda: quota_request.sign(object())
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_operation_has_no_method_name(self):
        op = sc_messages.QuotaOperation(consumerId=_TEST_CONSUMER_ID)
        testf = lambda: quota_request.sign(
            sc_messages.AllocateQuotaRequest(allocateOperation=op))
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_operation_has_no_consumer_id(self):
        op = sc_messages.QuotaOperation(methodName=_TEST_OP_NAME)
        testf = lambda: quota_request.sign(
            sc_messages.AllocateQuotaRequest(allocateOperation=op))
        expect(testf).to(raise_error(ValueError))

    def test_should_sign_a_valid_quota_request(self):
        quota_request.sign(self.test_quota_request)

    def test_should_change_signature_when_labels_are_added(self):
        without_labels = quota_request.sign(self.test_quota_request)
        self.test_op.labels = encoding.PyValueToMessage(
            sc_messages.QuotaOperation.LabelsValue, {
                u'key1': u'value1',
                u'key2': u'value2'})
        with_labels = quota_request.sign(self.test_quota_request)
        expect(with_labels).not_to(equal(without_labels))

    def test_should_change_signature_when_metric_values_are_added(self):
        without_mvs = quota_request.sign(self.test_quota_request)
        self.test_op.quotaMetrics = [
            sc_messages.MetricValueSet(
                metricName=u'a_float',
                metricValues=[
                    metric_value.create(
                        labels={
                            u'key1': u'value1',
                            u'key2': u'value2'},
                        doubleValue=1.1,
                    ),
                ]
            )
        ]
        with_mvs = quota_request.sign(self.test_quota_request)
        expect(with_mvs).not_to(equal(without_mvs))


class TestAggregatorQuota(unittest2.TestCase):
    SERVICE_NAME = u'service.quota'
    FAKE_OPERATION_ID = u'service.general.quota'

    def setUp(self):
        self.timer = _DateTimeTimer()
        self.agg = quota_request.Aggregator(
            self.SERVICE_NAME, caches.QuotaOptions())

    def test_should_fail_if_req_is_bad(self):
        testf = lambda: self.agg.allocate_quota(object())
        expect(testf).to(raise_error(ValueError))
        testf = lambda: self.agg.allocate_quota(None)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_service_name_does_not_match(self):
        req = _make_test_request(self.SERVICE_NAME + u'-will-not-match')
        testf = lambda: self.agg.allocate_quota(req)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_quota_request_is_missing(self):
        req = sc_messages.ServicecontrolServicesAllocateQuotaRequest(
            serviceName=self.SERVICE_NAME)
        testf = lambda: self.agg.allocate_quota(req)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_if_operation_is_missing(self):
        req = sc_messages.ServicecontrolServicesAllocateQuotaRequest(
            serviceName=self.SERVICE_NAME,
            allocateQuotaRequest=sc_messages.AllocateQuotaRequest())
        testf = lambda: self.agg.allocate_quota(req)
        expect(testf).to(raise_error(ValueError))

    def test_should_return_none_initially_as_req_is_not_cached(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        fake_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        agg = self.agg
        actual = agg.allocate_quota(req)
        expect(actual).to(equal(fake_response))


class TestAggregatorThatCannotCache(unittest2.TestCase):
    SERVICE_NAME = u'service.no_cache'
    FAKE_OPERATION_ID = u'service.no_cache.op_id'

    def setUp(self):
        # -ve num_entries means no cache is present
        self.agg = quota_request.Aggregator(
            self.SERVICE_NAME,
            caches.QuotaOptions(num_entries=-1))

    def test_should_not_cache_responses(self):
        req = _make_test_request(self.SERVICE_NAME)
        fake_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        agg = self.agg
        expect(agg.allocate_quota(req)).to(be_none)
        agg.add_response(req, fake_response)
        expect(agg.allocate_quota(req)).to(be_none)
        agg.clear()
        expect(agg.allocate_quota(req)).to(be_none)

    def test_should_have_empty_flush_response(self):
        expect(len(self.agg.flush())).to(equal(0))

    def test_should_have_none_as_flush_interval(self):
        expect(self.agg.flush_interval).to(be_none)



class TestCachingAggregator(unittest2.TestCase):
    SERVICE_NAME = u'service.with_cache'
    FAKE_OPERATION_ID = u'service.with_cache.op_id'

    def setUp(self):
        self.timer = _DateTimeTimer()
        self.expiration = datetime.timedelta(seconds=2)
        self.flush_interval = datetime.timedelta(seconds=1)
        options = caches.QuotaOptions(
            flush_interval=self.flush_interval,
            expiration=self.expiration)
        self.agg = quota_request.Aggregator(
            self.SERVICE_NAME, options, timer=self.timer)

    def test_should_have_flush_interval_as_flush_interval(self):
        expect(self.agg.flush_interval).to(equal(self.flush_interval))

    def test_should_cache_responses(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        temp_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        real_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID,
            quotaMetrics=[sc_messages.MetricValueSet(
                metricName=u'a_float',
                metricValues=[
                    metric_value.create(
                        labels={
                            u'key1': u'value1',
                            u'key2': u'value2'},
                        doubleValue=1.1,
                    ),
                ]
            )]
        )
        agg = self.agg
        expect(agg.allocate_quota(req)).to(equal(temp_response))
        agg.add_response(req, real_response)
        expect(agg.allocate_quota(req)).to(equal(real_response))

    def test_should_update_temp_response_with_actual(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        temp_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        real_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID,
            quotaMetrics=[sc_messages.MetricValueSet(
                metricName=u'a_float',
                metricValues=[
                    metric_value.create(
                        labels={
                            u'key1': u'value1',
                            u'key2': u'value2'},
                        doubleValue=1.1,
                    ),
                ]
            )]
        )
        agg = self.agg
        agg.allocate_quota(req)
        signature = quota_request.sign(req.allocateQuotaRequest)
        with agg._cache as cache:
            item = cache[signature]
            expect(item.response).to(equal(temp_response))
            expect(item.is_in_flight).to(be_true)
            agg.add_response(req, real_response)
            item = cache[signature]
            expect(item.response).to(equal(real_response))
            expect(item.is_in_flight).to(be_false)

    def test_request_aggregation(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        signature = quota_request.sign(req.allocateQuotaRequest)
        agg = self.agg
        agg.allocate_quota(req)
        with agg._cache as cache:
            item = cache[signature]
            expect(item._op_aggregator).to(be_none)
        agg.allocate_quota(req)
        agg.allocate_quota(req)
        with agg._out as out:
            expect(len(out)).to(equal(1))
        with agg._cache as cache:
            item = cache[signature]
            expect(item._op_aggregator).not_to(be_none)

    def test_aggregated_requests_should_be_sent_on_flush(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        signature = quota_request.sign(req.allocateQuotaRequest)
        agg = self.agg
        agg.allocate_quota(req)
        expect(len(agg.flush())).to(equal(1))  # initial request
        simple_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        agg.add_response(req, simple_response)
        agg.allocate_quota(req)
        agg.allocate_quota(req)
        expect(len(agg.flush())).to(equal(1))  # aggregated next two requests

    def test_expiration(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        temp_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        real_response = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID,
            quotaMetrics=[sc_messages.MetricValueSet(
                metricName=u'a_float',
                metricValues=[
                    metric_value.create(
                        labels={
                            u'key1': u'value1',
                            u'key2': u'value2'},
                        doubleValue=1.1,
                    ),
                ]
            )]
        )
        agg = self.agg
        agg.allocate_quota(req)
        assert len(agg.flush()) == 1
        agg.add_response(req, real_response)
        signature = quota_request.sign(req.allocateQuotaRequest)
        with agg._cache as cache, agg._out as out:
            assert len(out) == 0
            assert signature in cache
            self.timer.tick()
            self.timer.tick()
            assert len(agg.flush()) == 0
            assert len(out) == 0
            assert signature in cache
            self.timer.tick()  # expired at 3rd second
            assert len(agg.flush()) == 0
            assert len(out) == 0
            assert signature not in cache


class TestCacheItem(unittest2.TestCase):
    SERVICE_NAME = u'service.quota'
    FAKE_OPERATION_ID = u'service.general.quota'

    def test_request_aggregation(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        req = req.allocateQuotaRequest
        resp = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        item = quota_request.CachedItem(req, resp, self.SERVICE_NAME, None)
        expect(item._op_aggregator).to(be_none)
        with mock.patch.object(quota_request, 'QuotaOperationAggregator') as QOA:
            agg = QOA.return_value
            item.aggregate(req)
            expect(item._op_aggregator).to(be(agg))
            QOA.assert_called_once_with(req.allocateOperation)
            item.aggregate(req)
            agg.merge_operation.assert_called_once_with(req.allocateOperation)

    def test_request_extraction_no_aggregation(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        a_req = req.allocateQuotaRequest
        resp = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        item = quota_request.CachedItem(a_req, resp, self.SERVICE_NAME, None)
        expect(item._op_aggregator).to(be_none)
        expect(item.extract_request()).to(equal(req))

    def test_request_extraction_with_aggregation(self):
        req = _make_test_request(self.SERVICE_NAME, self.FAKE_OPERATION_ID)
        req = req.allocateQuotaRequest
        req.allocateOperation.quotaMetrics = [
            sc_messages.MetricValueSet(
                metricName=u'a_float',
                metricValues=[
                    metric_value.create(
                        labels={
                            u'key1': u'value1',
                            u'key2': u'value2'},
                        int64Value=12,
                    ),
                ]
            )
        ]

        resp = sc_messages.AllocateQuotaResponse(
            operationId=self.FAKE_OPERATION_ID)
        item = quota_request.CachedItem(req, resp, self.SERVICE_NAME, None)
        expect(item._op_aggregator).to(be_none)
        item.aggregate(req)
        item.aggregate(req)

        extracted = item.extract_request()
        op = extracted.allocateQuotaRequest.allocateOperation
        expect(op.quotaMetrics[0].metricValues[0].int64Value).to(equal(24))


_TEST_CONSUMER_ID = u'testConsumerID'
_TEST_OP_NAME = u'testOperationName'


def _make_test_request(service_name, operation_id=None, importance=None):
    if importance is None:
        importance = sc_messages.Operation.ImportanceValueValuesEnum.LOW
    op = sc_messages.QuotaOperation(
        consumerId=_TEST_CONSUMER_ID,
        methodName=_TEST_OP_NAME,
        operationId=operation_id,
    )
    quota_request = sc_messages.AllocateQuotaRequest(allocateOperation=op)
    return sc_messages.ServicecontrolServicesAllocateQuotaRequest(
        serviceName=service_name,
        allocateQuotaRequest=quota_request)


_WANTED_USER_AGENT = label_descriptor.USER_AGENT
_WANTED_SERVICE_AGENT = label_descriptor.SERVICE_AGENT
_START_OF_EPOCH = timestamp.to_rfc3339(datetime.datetime(1970, 1, 1, 0, 0, 0))
_TEST_SERVICE_NAME = u'a_service_name'
_INFO_TESTS = [
    (quota_request.Info(
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        referer=u'a_referer',
        service_name=_TEST_SERVICE_NAME),
     sc_messages.QuotaOperation(
         labels = encoding.PyValueToMessage(
             sc_messages.QuotaOperation.LabelsValue, {
                 u'servicecontrol.googleapis.com/referer': u'a_referer',
             }),
         operationId=u'an_op_id',
         methodName=u'an_op_name',
         quotaMode=sc_messages.QuotaOperation.QuotaModeValueValuesEnum.BEST_EFFORT)),
    (quota_request.Info(
        android_cert_fingerprint=u'an_android_cert_fingerprint',
        android_package_name=u'an_android_package_name',
        api_key=u'an_api_key',
        api_key_valid=True,
        ios_bundle_id=u'an_ios_bundle_id',
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        referer=u'a_referer',
        service_name=_TEST_SERVICE_NAME),
     sc_messages.QuotaOperation(
         consumerId=u'api_key:an_api_key',
         labels = encoding.PyValueToMessage(
             sc_messages.QuotaOperation.LabelsValue, {
                 u'servicecontrol.googleapis.com/referer': u'a_referer',
             }),
         operationId=u'an_op_id',
         methodName=u'an_op_name',
         quotaMode=sc_messages.QuotaOperation.QuotaModeValueValuesEnum.BEST_EFFORT)),
    (quota_request.Info(
        api_key=u'an_api_key',
        api_key_valid=False,
        client_ip=u'127.0.0.1',
        consumer_project_id=u'project_id',
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        referer=u'a_referer',
        service_name=_TEST_SERVICE_NAME),
     sc_messages.QuotaOperation(
         consumerId=u'project:project_id',
         labels = encoding.PyValueToMessage(
             sc_messages.QuotaOperation.LabelsValue, {
                 u'servicecontrol.googleapis.com/caller_ip': u'127.0.0.1',
                 u'servicecontrol.googleapis.com/referer': u'a_referer',
             }),
         operationId=u'an_op_id',
         methodName=u'an_op_name',
         quotaMode=sc_messages.QuotaOperation.QuotaModeValueValuesEnum.BEST_EFFORT)),
]
_INCOMPLETE_INFO_TESTS = [
    quota_request.Info(
        operation_name=u'an_op_name',
        service_name=_TEST_SERVICE_NAME),
    quota_request.Info(
        operation_id=u'an_op_id',
        service_name=_TEST_SERVICE_NAME),
    quota_request.Info(
        operation_id=u'an_op_id',
        operation_name=u'an_op_name')
]

KEYGETTER = attrgetter(u'key')


class TestInfo(unittest2.TestCase):
    def test_should_construct_with_no_args(self):
        expect(quota_request.Info()).not_to(be_none)

    def test_should_convert_using_as_quota_request(self):
        timer = _DateTimeTimer()
        for info, want in _INFO_TESTS:
            got = info.as_allocate_quota_request(timer=timer)
            # These additional properties have no well-defined order, so sort them.
            got.allocateQuotaRequest.allocateOperation.labels.additionalProperties.sort(key=KEYGETTER)
            want.labels.additionalProperties.sort(key=KEYGETTER)
            expect(got.allocateQuotaRequest.allocateOperation).to(equal(want))
            expect(got.serviceName).to(equal(_TEST_SERVICE_NAME))

    def test_should_fail_as_quota_request_on_incomplete_info(self):
        timer = _DateTimeTimer()
        for info in _INCOMPLETE_INFO_TESTS:
            testf = lambda: info.as_allocate_quota_request(timer=timer)
            expect(testf).to(raise_error(ValueError))


class TestConvertResponse(unittest2.TestCase):
    PROJECT_ID = u'test_convert_response'

    def test_should_be_ok_with_no_errors(self):
        code, message = quota_request.convert_response(
            sc_messages.AllocateQuotaResponse(), self.PROJECT_ID)
        expect(code).to(equal(httplib.OK))
        expect(message).to(equal(u''))

    def test_should_include_project_id_in_error_text_when_needed(self):
        resp = sc_messages.AllocateQuotaResponse(
            allocateErrors = [
                sc_messages.QuotaError(
                    code=sc_messages.QuotaError.CodeValueValuesEnum.PROJECT_DELETED)
            ]
        )
        code, got = quota_request.convert_response(resp, self.PROJECT_ID)
        want = u'Project %s has been deleted' % (self.PROJECT_ID,)
        expect(code).to(equal(httplib.FORBIDDEN))
        expect(got).to(equal(want))

    def test_should_include_detail_in_error_text_when_needed(self):
        detail = u'details, details, details'
        resp = sc_messages.AllocateQuotaResponse(
            allocateErrors = [
                sc_messages.QuotaError(
                    code=sc_messages.QuotaError.CodeValueValuesEnum.OUT_OF_RANGE,
                    description=detail)
            ]
        )
        code, got = quota_request.convert_response(resp, self.PROJECT_ID)
        expect(code).to(equal(httplib.INTERNAL_SERVER_ERROR))
        assert got.endswith(detail)


class _DateTimeTimer(object):
    def __init__(self, auto=False):
        self.auto = auto
        self.time = datetime.datetime.utcfromtimestamp(0)

    def __call__(self):
        if self.auto:
            self.tick()
        return self.time

    def tick(self):
        self.time += datetime.timedelta(seconds=1)



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generated client library for servicecontrol version v1."""
# NOTE: This file is autogenerated and should not be edited by hand.
from apitools.base.py import base_api
import servicecontrol_v1_messages as messages


class ServicecontrolV1(base_api.BaseApiClient):
  """Generated client library for service servicecontrol version v1."""

  MESSAGES_MODULE = messages
  BASE_URL = u'https://servicecontrol.googleapis.com/'

  _PACKAGE = u'servicecontrol'
  _SCOPES = [u'https://www.googleapis.com/auth/cloud-platform', u'https://www.googleapis.com/auth/servicecontrol']
  _VERSION = u'v1'
  _CLIENT_ID = '1042881264118.apps.googleusercontent.com'
  _CLIENT_SECRET = 'x_Tw5K8nnjoRAqULM9PFAC2b'
  _USER_AGENT = 'x_Tw5K8nnjoRAqULM9PFAC2b'
  _CLIENT_CLASS_NAME = u'ServicecontrolV1'
  _URL_VERSION = u'v1'
  _API_KEY = None

  def __init__(self, url='', credentials=None,
               get_credentials=True, http=None, model=None,
               log_request=False, log_response=False,
               credentials_args=None, default_global_params=None,
               additional_http_headers=None):
    """Create a new servicecontrol handle."""
    url = url or self.BASE_URL
    super(ServicecontrolV1, self).__init__(
        url, credentials=credentials,
        get_credentials=get_credentials, http=http, model=model,
        log_request=log_request, log_response=log_response,
        credentials_args=credentials_args,
        default_global_params=default_global_params,
        additional_http_headers=additional_http_headers)
    self.services = self.ServicesService(self)

  class ServicesService(base_api.BaseApiService):
    """Service class for the services resource."""

    _NAME = u'services'

    def __init__(self, client):
      super(ServicecontrolV1.ServicesService, self).__init__(client)
      self._upload_configs = {
          }

    def AllocateQuota(self, request, global_params=None):
      """Attempts to allocate quota for the specified consumer. It should be called.
before the operation is executed.

This method requires the `servicemanagement.services.quota`
permission on the specified service. For more information, see
[Google Cloud IAM](https://cloud.google.com/iam).

**NOTE:** the client code **must** fail-open if the server returns one
of the following quota errors:
-   `PROJECT_STATUS_UNAVAILABLE`
-   `SERVICE_STATUS_UNAVAILABLE`
-   `BILLING_STATUS_UNAVAILABLE`
-   `QUOTA_SYSTEM_UNAVAILABLE`

The server may inject above errors to prohibit any hard dependency
on the quota system.

      Args:
        request: (ServicecontrolServicesAllocateQuotaRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (AllocateQuotaResponse) The response message.
      """
      config = self.GetMethodConfig('AllocateQuota')
      return self._RunMethod(
          config, request, global_params=global_params)

    AllocateQuota.method_config = lambda: base_api.ApiMethodInfo(
        http_method=u'POST',
        method_id=u'servicecontrol.services.allocateQuota',
        ordered_params=[u'serviceName'],
        path_params=[u'serviceName'],
        query_params=[],
        relative_path=u'v1/services/{serviceName}:allocateQuota',
        request_field=u'allocateQuotaRequest',
        request_type_name=u'ServicecontrolServicesAllocateQuotaRequest',
        response_type_name=u'AllocateQuotaResponse',
        supports_download=False,
    )

    def Check(self, request, global_params=None):
      """Checks an operation with Google Service Control to decide whether.
the given operation should proceed. It should be called before the
operation is executed.

If feasible, the client should cache the check results and reuse them for
60 seconds. In case of server errors, the client can rely on the cached
results for longer time.

NOTE: the CheckRequest has the size limit of 64KB.

This method requires the `servicemanagement.services.check` permission
on the specified service. For more information, see
[Google Cloud IAM](https://cloud.google.com/iam).

      Args:
        request: (ServicecontrolServicesCheckRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (CheckResponse) The response message.
      """
      config = self.GetMethodConfig('Check')
      return self._RunMethod(
          config, request, global_params=global_params)

    Check.method_config = lambda: base_api.ApiMethodInfo(
        http_method=u'POST',
        method_id=u'servicecontrol.services.check',
        ordered_params=[u'serviceName'],
        path_params=[u'serviceName'],
        query_params=[],
        relative_path=u'v1/services/{serviceName}:check',
        request_field=u'checkRequest',
        request_type_name=u'ServicecontrolServicesCheckRequest',
        response_type_name=u'CheckResponse',
        supports_download=False,
    )

    def EndReconciliation(self, request, global_params=None):
      """Signals the quota controller that service ends the ongoing usage.
reconciliation.

This method requires the `servicemanagement.services.quota`
permission on the specified service. For more information, see
[Google Cloud IAM](https://cloud.google.com/iam).

      Args:
        request: (ServicecontrolServicesEndReconciliationRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (EndReconciliationResponse) The response message.
      """
      config = self.GetMethodConfig('EndReconciliation')
      return self._RunMethod(
          config, request, global_params=global_params)

    EndReconciliation.method_config = lambda: base_api.ApiMethodInfo(
        http_method=u'POST',
        method_id=u'servicecontrol.services.endReconciliation',
        ordered_params=[u'serviceName'],
        path_params=[u'serviceName'],
        query_params=[],
        relative_path=u'v1/services/{serviceName}:endReconciliation',
        request_field=u'endReconciliationRequest',
        request_type_name=u'ServicecontrolServicesEndReconciliationRequest',
        response_type_name=u'EndReconciliationResponse',
        supports_download=False,
    )

    def ReleaseQuota(self, request, global_params=None):
      """Releases previously allocated quota done through AllocateQuota method.

This method requires the `servicemanagement.services.quota`
permission on the specified service. For more information, see
[Google Cloud IAM](https://cloud.google.com/iam).

**NOTE:** the client code **must** fail-open if the server returns one
of the following quota errors:
-   `PROJECT_STATUS_UNAVAILABLE`
-   `SERVICE_STATUS_UNAVAILABLE`
-   `BILLING_STATUS_UNAVAILABLE`
-   `QUOTA_SYSTEM_UNAVAILABLE`

The server may inject above errors to prohibit any hard dependency
on the quota system.

      Args:
        request: (ServicecontrolServicesReleaseQuotaRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ReleaseQuotaResponse) The response message.
      """
      config = self.GetMethodConfig('ReleaseQuota')
      return self._RunMethod(
          config, request, global_params=global_params)

    ReleaseQuota.method_config = lambda: base_api.ApiMethodInfo(
        http_method=u'POST',
        method_id=u'servicecontrol.services.releaseQuota',
        ordered_params=[u'serviceName'],
        path_params=[u'serviceName'],
        query_params=[],
        relative_path=u'v1/services/{serviceName}:releaseQuota',
        request_field=u'releaseQuotaRequest',
        request_type_name=u'ServicecontrolServicesReleaseQuotaRequest',
        response_type_name=u'ReleaseQuotaResponse',
        supports_download=False,
    )

    def Report(self, request, global_params=None):
      """Reports operation results to Google Service Control, such as logs and.
metrics. It should be called after an operation is completed.

If feasible, the client should aggregate reporting data for up to 5
seconds to reduce API traffic. Limiting aggregation to 5 seconds is to
reduce data loss during client crashes. Clients should carefully choose
the aggregation time window to avoid data loss risk more than 0.01%
for business and compliance reasons.

NOTE: the ReportRequest has the size limit of 1MB.

This method requires the `servicemanagement.services.report` permission
on the specified service. For more information, see
[Google Cloud IAM](https://cloud.google.com/iam).

      Args:
        request: (ServicecontrolServicesReportRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ReportResponse) The response message.
      """
      config = self.GetMethodConfig('Report')
      return self._RunMethod(
          config, request, global_params=global_params)

    Report.method_config = lambda: base_api.ApiMethodInfo(
        http_method=u'POST',
        method_id=u'servicecontrol.services.report',
        ordered_params=[u'serviceName'],
        path_params=[u'serviceName'],
        query_params=[],
        relative_path=u'v1/services/{serviceName}:report',
        request_field=u'reportRequest',
        request_type_name=u'ServicecontrolServicesReportRequest',
        response_type_name=u'ReportResponse',
        supports_download=False,
    )

    def StartReconciliation(self, request, global_params=None):
      """Unlike rate quota, allocation quota does not get refilled periodically.
So, it is possible that the quota usage as seen by the service differs from
what the One Platform considers the usage is. This is expected to happen
only rarely, but over time this can accumulate. Services can invoke
StartReconciliation and EndReconciliation to correct this usage drift, as
described below:
1. Service sends StartReconciliation with a timestamp in future for each
   metric that needs to be reconciled. The timestamp being in future allows
   to account for in-flight AllocateQuota and ReleaseQuota requests for the
   same metric.
2. One Platform records this timestamp and starts tracking subsequent
   AllocateQuota and ReleaseQuota requests until EndReconciliation is
   called.
3. At or after the time specified in the StartReconciliation, service
   sends EndReconciliation with the usage that needs to be reconciled to.
4. One Platform adjusts its own record of usage for that metric to the
   value specified in EndReconciliation by taking in to account any
   allocation or release between StartReconciliation and EndReconciliation.

Signals the quota controller that the service wants to perform a usage
reconciliation as specified in the request.

This method requires the `servicemanagement.services.quota`
permission on the specified service. For more information, see
[Google Cloud IAM](https://cloud.google.com/iam).

      Args:
        request: (ServicecontrolServicesStartReconciliationRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (StartReconciliationResponse) The response message.
      """
      config = self.GetMethodConfig('StartReconciliation')
      return self._RunMethod(
          config, request, global_params=global_params)

    StartReconciliation.method_config = lambda: base_api.ApiMethodInfo(
        http_method=u'POST',
        method_id=u'servicecontrol.services.startReconciliation',
        ordered_params=[u'serviceName'],
        path_params=[u'serviceName'],
        query_params=[],
        relative_path=u'v1/services/{serviceName}:startReconciliation',
        request_field=u'startReconciliationRequest',
        request_type_name=u'ServicecontrolServicesStartReconciliationRequest',
        response_type_name=u'StartReconciliationResponse',
        supports_download=False,
    )



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
import httmock
import mock

from Crypto import PublicKey
from jwkest import jwk

from endpoints_management.auth import suppliers


class KeyUriSupplierTest(unittest.TestCase):

    def test_supply_issuer(self):
        issuer = u"https://issuer.com"
        jwks_uri = u"https://issuer.com/jwks/uri"
        configs = {issuer: suppliers.IssuerUriConfig(False, jwks_uri)}
        supplier = suppliers.KeyUriSupplier(configs)
        self.assertEquals(jwks_uri, supplier.supply(issuer))
        self.assertIsNone(supplier.supply(u"random-issuer"))

    def test_openid_discovery(self):
        jwks_uri = u"https://issuer.com/jwks/uri"
        @httmock.urlmatch(scheme=u"https", netloc=u"issuer.com",
                          path=u"/" + suppliers._OPEN_ID_CONFIG_PATH)
        def _mock_response(url, request):  # pylint: disable=unused-argument
            response = {u"jwks_uri": jwks_uri}
            return json.dumps(response)

        issuer = u"https://issuer.com"
        configs = {issuer: suppliers.IssuerUriConfig(True, None)}
        supplier = suppliers.KeyUriSupplier(configs)
        with httmock.HTTMock(_mock_response):
            self.assertEquals(jwks_uri, supplier.supply(issuer))

    def test_issuer_without_protocol(self):
        jwks_uri = u"https://issuer.com/jwks/uri"
        @httmock.urlmatch(scheme=u"https", netloc=u"issuer.com",
                          path=u"/" + suppliers._OPEN_ID_CONFIG_PATH)
        def _mock_response(url, request):  # pylint: disable=unused-argument
            response = {u"jwks_uri": jwks_uri}
            return json.dumps(response)

        # Specify an issuer without protocol to make sure the "https://" prefix is
        # added automatically.
        issuer = u"issuer.com"
        configs = {issuer: suppliers.IssuerUriConfig(True, None)}
        supplier = suppliers.KeyUriSupplier(configs)
        with httmock.HTTMock(_mock_response):
            self.assertEquals(jwks_uri, supplier.supply(issuer))

    def test_openid_discovery_with_bad_json(self):
        @httmock.urlmatch(scheme=u"https", netloc=u"issuer.com")
        def _mock_response_with_bad_json(url, request):  # pylint: disable=unused-argument
            return u"bad-json"

        issuer = u"https://issuer.com"
        configs = {issuer: suppliers.IssuerUriConfig(True, None)}
        supplier = suppliers.KeyUriSupplier(configs)
        with httmock.HTTMock(_mock_response_with_bad_json):
            with self.assertRaises(suppliers.UnauthenticatedException):
                supplier.supply(issuer)


class JwksSupplierTest(unittest.TestCase):
    _mock_timer = mock.MagicMock()

    def setUp(self):
        self._key_uri_supplier = mock.MagicMock()
        self._jwks_uri_supplier = suppliers.JwksSupplier(self._key_uri_supplier)

    def test_supply_with_unknown_issuer(self):
        self._key_uri_supplier.supply.return_value = None
        issuer = u"unknown-issuer"
        expected_message = u"Cannot find the `jwks_uri` for issuer " + issuer
        with self.assertRaisesRegexp(suppliers.UnauthenticatedException,
                                     expected_message):
            self._jwks_uri_supplier.supply(issuer)

    def test_supply_with_invalid_json_response(self):
        scheme = u"https"
        issuer = u"issuer.com"
        self._key_uri_supplier.supply.return_value = scheme + u"://" + issuer

        @httmock.urlmatch(scheme=scheme, netloc=issuer)
        def _mock_response_with_invalid_json(url, response):  # pylint: disable=unused-argument
            return u"invalid-json"

        with httmock.HTTMock(_mock_response_with_invalid_json):
            with self.assertRaises(suppliers.UnauthenticatedException):
                self._jwks_uri_supplier.supply(issuer)

    def test_supply_jwks(self):
        rsa_key = PublicKey.RSA.generate(2048)
        jwks = jwk.KEYS()
        jwks.wrap_add(rsa_key)

        scheme = u"https"
        issuer = u"issuer.com"
        self._key_uri_supplier.supply.return_value = scheme + u"://" + issuer

        @httmock.urlmatch(scheme=scheme, netloc=issuer)
        def _mock_response_with_jwks(url, response):  # pylint: disable=unused-argument
            return jwks.dump_jwks()

        with httmock.HTTMock(_mock_response_with_jwks):
            actual_jwks = self._jwks_uri_supplier.supply(issuer)
            self.assertEquals(1, len(actual_jwks))
            actual_key = actual_jwks[0].key
            self.assertEquals(rsa_key.n, actual_key.n)
            self.assertEquals(rsa_key.e, actual_key.e)

    def test_supply_jwks_with_x509_certificate(self):
        rsa_key = PublicKey.RSA.generate(2048)
        cert = rsa_key.publickey().exportKey(u"PEM")
        kid = u"rsa-cert"

        scheme = u"https"
        issuer = u"issuer.com"
        self._key_uri_supplier.supply.return_value = scheme + u"://" + issuer

        @httmock.urlmatch(scheme=scheme, netloc=issuer)
        def _mock_response_with_x509_certificates(url, response):  # pylint: disable=unused-argument
            return json.dumps({kid: cert.decode('ascii')})

        with httmock.HTTMock(_mock_response_with_x509_certificates):
            actual_jwks = self._jwks_uri_supplier.supply(issuer)
            self.assertEquals(1, len(actual_jwks))
            actual_key = actual_jwks[0].key

            self.assertEquals(kid, actual_jwks[0].kid)
            self.assertEquals(rsa_key.n, actual_key.n)
            self.assertEquals(rsa_key.e, actual_key.e)

    def test_supply_empty_x509_certificate(self):
        scheme = u"https"
        issuer = u"issuer.com"
        self._key_uri_supplier.supply.return_value = scheme + u"://" + issuer

        @httmock.urlmatch(scheme=scheme, netloc=issuer)
        def _mock_invalid_response(url, response):  # pylint: disable=unused-argument
            return json.dumps({u"kid": u"invlid-certificate"})

        with httmock.HTTMock(_mock_invalid_response):
            with self.assertRaises(suppliers.UnauthenticatedException):
                self._jwks_uri_supplier.supply(issuer)

    @mock.patch(u"time.time", _mock_timer)
    def test_supply_cached_jwks(self):
        JwksSupplierTest._mock_timer.return_value = 10
        rsa_key = PublicKey.RSA.generate(2048)
        jwks = jwk.KEYS()
        jwks.wrap_add(rsa_key)

        scheme = u"https"
        issuer = u"issuer.com"
        self._key_uri_supplier.supply.return_value = scheme + u"://" + issuer

        @httmock.urlmatch(scheme=scheme, netloc=issuer)
        def _mock_response_with_jwks(url, response):  # pylint: disable=unused-argument
            return jwks.dump_jwks()

        with httmock.HTTMock(_mock_response_with_jwks):
            self.assertEqual(1, len(self._jwks_uri_supplier.supply(issuer)))

            # Add an additional key to the JWKS to be returned by the HTTP request.
            jwks.wrap_add(PublicKey.RSA.generate(2048))

            # Forward the clock by 1 second. The JWKS should remain cached.
            JwksSupplierTest._mock_timer.return_value += 1
            self._jwks_uri_supplier.supply(issuer)
            self.assertEqual(1, len(self._jwks_uri_supplier.supply(issuer)))

            # Forward the clock by 5 minutes. The cache entry should have expired so
            # the returned JWKS should be the updated one with two keys.
            JwksSupplierTest._mock_timer.return_value += 5 * 60
            self._jwks_uri_supplier.supply(issuer)
            self.assertEqual(2, len(self._jwks_uri_supplier.supply(issuer)))



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import mock
import json
import time
import unittest

from Crypto import PublicKey
from jwkest import ecc
from jwkest import jwk
from jwkest import jws
from test import token_utils

from endpoints_management.auth import suppliers
from endpoints_management.auth import tokens


class AuthenticatorTest(unittest.TestCase):
    _ec_kid = u"ec-key-id"
    _rsa_kid = u"rsa-key-id"

    _mock_timer = mock.MagicMock()

    def setUp(self):
        ec_jwk = jwk.ECKey(use=u"sig").load_key(ecc.P256)
        ec_jwk.kid = self._ec_kid

        rsa_key = jwk.RSAKey(use=u"sig").load_key(PublicKey.RSA.generate(1024))
        rsa_key.kid = self._rsa_kid

        jwks = jwk.KEYS()
        jwks._keys.append(ec_jwk)
        jwks._keys.append(rsa_key)

        self._issuers_to_provider_ids = {}
        self._jwks_supplier = mock.MagicMock()
        self._authenticator = tokens.Authenticator(self._issuers_to_provider_ids,
                                                   self._jwks_supplier)
        self._jwks = jwks
        self._jwks_supplier.supply.return_value = self._jwks

        self._method_info = mock.MagicMock()
        self._service_name = u"service.name.com"

        self._jwt_claims = {
            u"aud": [u"first.com", u"second.com"],
            u"email": u"someone@email.com",
            u"exp": int(time.time()) + 10,
            u"iss": u"https://issuer.com",
            u"sub": u"subject-id"}

    def test_get_jwt_claims(self):
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys,
                                                     kid=self._ec_kid)
        actual_jwt_claims = self._authenticator.get_jwt_claims(auth_token)
        self.assertEqual(self._jwt_claims, actual_jwt_claims)

    def test_get_jwt_claims_without_kid(self):
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys)
        actual_jwt_claims = self._authenticator.get_jwt_claims(auth_token)
        self.assertEqual(self._jwt_claims, actual_jwt_claims)

    def test_required_claims(self):
        def assert_missing_claim_raise_exception(claim_name):
            jwt_claims = copy.deepcopy(self._jwt_claims)
            del jwt_claims[claim_name]
            auth_token = token_utils.generate_auth_token(jwt_claims,
                                                         self._jwks._keys,
                                             kid=self._ec_kid)
            with self.assertRaisesRegexp(suppliers.UnauthenticatedException,
                                         u'Missing "%s" claim' % claim_name):
                self._authenticator.get_jwt_claims(auth_token)

        assert_missing_claim_raise_exception(u"aud")
        assert_missing_claim_raise_exception(u"exp")
        assert_missing_claim_raise_exception(u"sub")
        assert_missing_claim_raise_exception(u"iss")

    @mock.patch(u"time.time", _mock_timer)
    def test_get_jwt_claims_via_caching(self):
        AuthenticatorTest._mock_timer.return_value = 10

        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys)
        # Populate the decoded result into cache.
        self._authenticator.get_jwt_claims(auth_token)

        # Reset the returned JWKS so the signature verification will fail next
        # time.
        self._jwks_supplier.supply.return_value = jwk.KEYS()

        # Forword time by 10 seconds.
        AuthenticatorTest._mock_timer.return_value += 10
        # This call should succeed since the auth_token is cached.
        self._authenticator.get_jwt_claims(auth_token)

        # Forword time by 5 minutes.
        AuthenticatorTest._mock_timer.return_value += 5 * 60
        # This call should fail since the cache expires and it needs to re-decode
        # the auth token with a different key set.
        with self.assertRaises(suppliers.UnauthenticatedException):
            self._authenticator.get_jwt_claims(auth_token)

    def test_auth_token_cache_capacity(self):
        authenticator = tokens.Authenticator({}, self._jwks_supplier, cache_capacity=2)

        self._jwt_claims[u"email"] = u"1@email.com"
        auth_token1 = token_utils.generate_auth_token(self._jwt_claims,
                                                      self._jwks._keys)
        self._jwt_claims[u"email"] = u"2@email.com"
        auth_token2 = token_utils.generate_auth_token(self._jwt_claims,
                                                      self._jwks._keys)

        # Populate the decoded result into cache.
        authenticator.get_jwt_claims(auth_token1)
        authenticator.get_jwt_claims(auth_token2)

        # Reset the returned JWKS so the signature verification will fail next
        # time.
        new_ec_jwk = jwk.ECKey(use=u"sig").load_key(ecc.P256)
        new_ec_jwk.kid = self._ec_kid
        new_jwks = jwk.KEYS()
        new_jwks._keys.append(new_ec_jwk)
        self._jwks_supplier.supply.return_value = new_jwks

        # Verify the following calls still succeed since the auth tokens are
        # cached.
        authenticator.get_jwt_claims(auth_token1)
        authenticator.get_jwt_claims(auth_token2)

        # Populate a third auth token into the cache.
        self._jwt_claims[u"email"] = u"3@email.com"
        auth_token3 = token_utils.generate_auth_token(self._jwt_claims,
                                                      new_jwks._keys)
        authenticator.get_jwt_claims(auth_token3)

        # Make sure the first auth token is evicted from the cache since the cache
        # is full.
        with self.assertRaises(suppliers.UnauthenticatedException):
            authenticator.get_jwt_claims(auth_token1)

    def test_verify_fails(self):
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys,
                                         kid=self._ec_kid)

        # Let the _jwks_supplier return a different key than the one we use to sign
        # the JWT.
        new_jwk = jwk.ECKey(use=u"sig").load_key(ecc.P256)
        new_jwks = jwk.KEYS()
        new_jwks._keys.append(new_jwk)
        self._jwks_supplier.supply.return_value = new_jwks

        with self.assertRaises(suppliers.UnauthenticatedException):
            self._authenticator.get_jwt_claims(auth_token)

    def test_authenticate_successfully(self):
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys,
                                                     kid=self._ec_kid)
        self._method_info.get_allowed_audiences.return_value = [u"first.com"]
        self._issuers_to_provider_ids[self._jwt_claims[u"iss"]] = u"provider-id"
        actual_user_info = self._authenticator.authenticate(auth_token,
                                                            self._method_info,
                                                            u"service.name.com")
        self.assert_user_info(actual_user_info, self._jwt_claims[u"aud"],
                              self._jwt_claims[u"email"], self._jwt_claims[u"sub"],
                              self._jwt_claims[u"iss"])

    def test_authenticate_with_single_audience(self):
        aud = u"first.aud.com"
        self._jwt_claims[u"aud"] = aud
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys,
                                                     kid=self._ec_kid)
        self._issuers_to_provider_ids[self._jwt_claims[u"iss"]] = u"provider-id"
        actual_user_info = self._authenticator.authenticate(auth_token,
                                                            self._method_info, aud)
        self.assertEqual([aud], actual_user_info.audiences)

    def test_authenticate_with_malformed_claims(self):
        def assert_malformed_time_claim_raises_exception(claim_name, expiration):
            jwt_claims = copy.deepcopy(self._jwt_claims)
            jwt_claims[claim_name] = expiration
            auth_token = token_utils.generate_auth_token(jwt_claims,
                                                         self._jwks._keys)
            message = u'Malformed claim: "%s" must be an integer' % claim_name
            with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
                self._authenticator.authenticate(auth_token, self._method_info,
                                                 u"service.name")

        assert_malformed_time_claim_raises_exception(u"exp", u"1")
        assert_malformed_time_claim_raises_exception(u"exp", 1.1)
        assert_malformed_time_claim_raises_exception(u"exp", [1])
        assert_malformed_time_claim_raises_exception(u"nbf", u"1")
        assert_malformed_time_claim_raises_exception(u"nbf", 1.1)
        assert_malformed_time_claim_raises_exception(u"nbf", [1])

    def test_authenticate_with_expired_auth_token(self):
        self._jwt_claims[u"exp"] = long(time.time() - 10)
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys)
        message = u"The auth token has already expired"
        with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
            self._authenticator.authenticate(auth_token,
                                             self._method_info,
                                             u"service.name")

    def test_authenticate_with_nbf_claim(self):
        # Set the "nbf" claim to some time in the future.
        self._jwt_claims[u"nbf"] = long(time.time() + 5)
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys)
        message = u'Current time is less than the "nbf" time'
        with self.assertRaisesRegexp(suppliers.UnauthenticatedException, message):
            self._authenticator.authenticate(auth_token, self._method_info,
                                             u"service.name")

    def test_authenticate_with_service_name_as_audience(self):
        self._jwt_claims[u"aud"].append(self._service_name)
        self._issuers_to_provider_ids[self._jwt_claims[u"iss"]] = u"provider-id"
        self._method_info.get_allowed_audiences.return_value = []
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys,
                                         kid=self._ec_kid)
        actual_user_info = self._authenticator.authenticate(auth_token,
                                                            self._method_info,
                                                            self._service_name)
        self.assert_user_info(actual_user_info, self._jwt_claims[u"aud"],
                              self._jwt_claims[u"email"], self._jwt_claims[u"sub"],
                              self._jwt_claims[u"iss"])

    def test_authenticate_with_disallowed_provider_id(self):
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys,
                                         kid=self._ec_kid)
        self._method_info.is_provider_allowed.return_value = False
        self._issuers_to_provider_ids[self._jwt_claims[u"iss"]] = u"id"
        with self.assertRaisesRegexp(suppliers.UnauthenticatedException,
                                     u"The requested method does not allow provider "
                                     u"id: id"):
            self._authenticator.authenticate(auth_token, self._method_info,
                                             self._service_name)

    def test_authenticate_with_disallowed_audiences(self):
        auth_token = token_utils.generate_auth_token(self._jwt_claims,
                                                     self._jwks._keys,
                                                     kid=self._ec_kid)
        self._method_info.get_allowed_audiences.return_value = []
        self._issuers_to_provider_ids[self._jwt_claims[u"iss"]] = u"project-id"
        with self.assertRaisesRegexp(suppliers.UnauthenticatedException,
                                     u"Audiences not allowed"):
            self._authenticator.authenticate(auth_token, self._method_info,
                                             self._service_name)

    def test_unicode_decode_error(self):
        auth_token = u"ya29.CjA8A3Hrca1hCCvRg69U3Tg85CG5pRqZj7gOJUsicpRafWAW63zvg6a0ZM6wZ5mJwM0"
        with self.assertRaisesRegexp(suppliers.UnauthenticatedException,
                                     u"Cannot decode the auth token"):
            self._authenticator.authenticate(auth_token, None, None)

    def assert_user_info(self, actual_user_info, audiences, email, subject_id,
                         issuer):
        self.assertEqual(audiences, actual_user_info.audiences)
        self.assertEqual(email, actual_user_info.email)
        self.assertEqual(subject_id, actual_user_info.subject_id)
        self.assertEqual(issuer, actual_user_info.issuer)
